/*============================================================================
  File:     RecoverDroppedTableFromLog.sql

  Summary:  How do you recover from a dropped table?
                This is easier but not all "damage" is
                this easy to find?! 
                  
  Date:     October 2011

  SQL Server Version: 2005/2008/R2
------------------------------------------------------------------------------
  Written by SQLskills.com

  For more scripts and sample code, check out 
    http://www.SQLskills.com

  This script is intended only as a supplement to demos and lectures
  written/presented by SQLskills.com  
  
  THIS CODE AND INFORMATION ARE PROVIDED "AS IS" WITHOUT WARRANTY OF 
  ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED 
  TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
  PARTICULAR PURPOSE.
============================================================================*/

------------------------------------------------------------------------------
-- ** STEP 1 Begin (cleanup) **
------------------------------------------------------------------------------
-- Delete backup history for SalesDB database
USE msdb;
GO

EXEC sp_delete_database_backuphistory 'SalesDB';
GO

-- Create a new full backup to use as the base for restore operations
USE master;
GO

BACKUP DATABASE SalesDB 
	TO DISK = N'C:\SQLskills\SalesDBBackup.bak'
	WITH NAME = N'Full Database Backup',
	DESCRIPTION = 'Starting point for recovery',
	INIT,
	STATS = 10;
GO

sp_helpfile
go
------------------------------------------------------------------------------
-- ** STEP 1 End **
------------------------------------------------------------------------------


------------------------------------------------------------------------------
-- ** STEP 2 Begin **
------------------------------------------------------------------------------
USE SalesDB;
GO

-- At restore:
-- Sales Row Count: 6,715,221
--     Max SalesID: 6,737,977

-- Kick off users

SELECT count(*) AS [RowCount]
        , max(SalesID) AS [MaxSalesID]
	FROM Sales 
	WITH (READUNCOMMITTED);
GO

SELECT getdate() AS "Date";
GO

-- Saved Values: 
--RowCount	MaxSalesID
--6720417	6743173

-- Saved "Known good" Date/time: 2011-10-14 14:44:55.763
------------------------------------------------------------------------------
-- ** STEP 2 End **
------------------------------------------------------------------------------


------------------------------------------------------------------------------
-- ** STEP 3 Begin **
------------------------------------------------------------------------------
-- Simulate a user dropping the unprotected table.
DROP TABLE Sales;
GO

-- TIME: 2:47pm
------------------------------------------------------------------------------
-- ** STEP 3 End **
------------------------------------------------------------------------------


------------------------------------------------------------------------------
-- ** STEP 4 Begin **
------------------------------------------------------------------------------
-- What are your options?
-- Examine the database file structure.
EXEC sp_helpfile;
GO

-- With one mdf/ldf - limited [are you options!]
------------------------------------------------------------------------------
-- ** STEP 4 End **
------------------------------------------------------------------------------


------------------------------------------------------------------------------
-- ** STEP 5 Begin **
------------------------------------------------------------------------------
-- Take the database offline to regular users. 
-- This is debatable depending on how the app will behave...
USE master;
GO

ALTER DATABASE SalesDB
	SET RESTRICTED_USER
	WITH ROLLBACK IMMEDIATE;
GO
------------------------------------------------------------------------------
-- ** STEP 5 End **
------------------------------------------------------------------------------


------------------------------------------------------------------------------
-- ** STEP 6 Begin **
------------------------------------------------------------------------------
-- Backup the log (if database inaccessible use no_truncate)
BACKUP LOG SalesDB 
	TO DISK = N'C:\SQLskills\SalesDBBackup.bak'
	WITH NAME = N'Transaction Log Backup',
	DESCRIPTION = 'Getting everything to current point in time.', 
	STATS = 10;
GO
------------------------------------------------------------------------------
-- ** STEP 6 End **
------------------------------------------------------------------------------


------------------------------------------------------------------------------
-- ** STEP 7 Begin **
------------------------------------------------------------------------------
-- See what backups we have
RESTORE HEADERONLY FROM
	DISK = N'C:\SQLskills\SalesDBBackup.bak';
GO

-- Query msdb:
SELECT Backup_Start_Date, 
	[Name],
	[Description],
	First_LSN, 
	Last_LSN, 
	Backup_Finish_Date, 
	* -- OR Just *
FROM msdb.dbo.backupset AS s
    JOIN msdb.dbo.backupmediafamily AS m
        ON s.media_set_id = m.media_set_id
WHERE database_name = 'SalesDB'
ORDER BY 1 ASC

-- Use standard reports:
--	Standard Reports: Backup and Restore Events

-- Use 3rd party software?
------------------------------------------------------------------------------
-- ** STEP 7 End **
------------------------------------------------------------------------------


------------------------------------------------------------------------------
-- ** STEP 8 Begin **
------------------------------------------------------------------------------
-- How can you find when a table was dropped?
-- Use a standard report (default trace)
--  Right click on DB name
--      Reports | Standard Reports | Schema Changes History
--          Gives you the time (to the second)

-- Use fn_dblog
USE SalesDB
go

SELECT [Transaction Id], [Begin Time], *
 FROM ::fn_dblog (NULL, NULL)
 WHERE [Transaction Name] = 'DROPOBJ'
 GO

-- What time?
-- Standard report time: 2:47:18
-- fn_dblog time: 2011/10/14 14:47:18:107
USE master;
GO

RESTORE DATABASE SalesDB
	FROM DISK = N'C:\SQLskills\SalesDBBackup.bak'
	WITH FILE = 1 --, REPLACE -- SQL doesn't allow you to restore over a database where the transaction log (the "tail" of the log) has not been backed up (unless you force it to with REPLACE)
	, STANDBY = N'C:\SQLskills\SalesDB_UNDO.bak'	
	, STOPAT = '2011/10/14 14:47:18' -- Do we know a point in time where things were good?
	, STATS = 10;
GO

RESTORE LOG SalesDB
	FROM DISK = N'C:\SQLskills\SalesDBBackup.bak'
	WITH FILE = 2
	 , STANDBY = N'C:\SQLskills\SalesDB_UNDO.bak'	
	 , STOPAT = '2011/10/14 14:47:18' -- Do we know a point in time where things were good?
	 , STATS = 10;
GO

-- Let's check our numbers:
SELECT count(*) AS [RowCount]
        , max(SalesID) AS [MaxSalesID]
	FROM SalesDB.dbo.Sales 
	WITH (READUNCOMMITTED);
GO

--RowCount	MaxSalesID
--6734010	6756766

-- Now - let's use the fn_dblog date/time:
RESTORE LOG SalesDB
	FROM DISK = N'C:\SQLskills\SalesDBBackup.bak'
	WITH FILE = 2
	 , STANDBY = N'C:\SQLskills\SalesDB_UNDO.bak'	
	 , STOPAT = ' 2011/10/14 14:47:18:107' -- Do we know a point in time where things were good?
	 , STATS = 10;
GO

-- Let's check our numbers:
SELECT count(*) AS [RowCount]
        , max(SalesID) AS [MaxSalesID]
	FROM SalesDB.dbo.Sales 
	WITH (READUNCOMMITTED);
GO

--RowCount	MaxSalesID
--6724316	6747072

RESTORE DATABASE SalesDB WITH RECOVERY, RESTRICTED_USER
go
------------------------------------------------------------------------------
-- ** STEP 8 End **
------------------------------------------------------------------------------


