/*============================================================================
  File:     InvestigatingTheTimeOfDisaster.sql

  Summary:  How do you find out what you don't know?
                Or, when it occurred. Or, who did it?
                  
  Date:     October 2011

  SQL Server Version: 2005/2008/R2
------------------------------------------------------------------------------
  Written by SQLskills.com

  For more scripts and sample code, check out 
    http://www.SQLskills.com

  This script is intended only as a supplement to demos and lectures
  written/presented by SQLskills.com  
  
  THIS CODE AND INFORMATION ARE PROVIDED "AS IS" WITHOUT WARRANTY OF 
  ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED 
  TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
  PARTICULAR PURPOSE.
============================================================================*/

------------------------------------------------------------------------------
-- ** STEP 1 Begin **
------------------------------------------------------------------------------
-- Delete backup history for SalesDB database
USE msdb;
GO

EXEC sp_delete_database_backuphistory 'SalesDB';
GO

-- Create a new full backup to use as the base for restore operations
USE master;
GO

BACKUP DATABASE SalesDB 
	TO DISK = N'C:\SQLskills\SalesDBBackup.bak'
	WITH NAME = N'Full Database Backup',
	DESCRIPTION = 'Starting point for recovery',
	INIT,
	STATS = 10;
GO
------------------------------------------------------------------------------
-- ** STEP 1 End **
------------------------------------------------------------------------------


------------------------------------------------------------------------------
-- ** STEP 2 Begin **
------------------------------------------------------------------------------
USE SalesDB;
GO

SELECT count(*) AS "Row Count"
	FROM Sales 
	WITH (READUNCOMMITTED);
GO

SELECT max(SalesID) AS "Max SalesID"
	FROM Sales 
	WITH (READUNCOMMITTED);
go

SELECT getdate() AS "Date";
GO
-- Saved Row Count: 6715924
-- Saved Max SalesID: 6738699
-- Saved "Known good" Date/time: 2010-07-08 14:00:44.473
------------------------------------------------------------------------------
-- ** STEP 2 End **
------------------------------------------------------------------------------


------------------------------------------------------------------------------
-- ** STEP 3 Begin **
------------------------------------------------------------------------------
-- Simulate a user dropping the unprotected table.
DROP TABLE Sales;
GO

-- Someone tell's you that data doesn't seem quite right?
-- Immediately make a note of the time!
-- Time Problem reported: HH:MM
------------------------------------------------------------------------------
-- ** STEP 3 End **
------------------------------------------------------------------------------


------------------------------------------------------------------------------
-- ** STEP 4 Begin **
------------------------------------------------------------------------------
-- What are your options?
-- Examine the database file structure.
EXEC sp_helpfile;
GO

-- With one mdf/ldf - limited [are you options!]
------------------------------------------------------------------------------
-- ** STEP 4 End **
------------------------------------------------------------------------------


------------------------------------------------------------------------------
-- ** STEP 5 Begin **
------------------------------------------------------------------------------
-- Take the database offline to regular users.
USE master;
GO

ALTER DATABASE SalesDB
	SET RESTRICTED_USER
	WITH ROLLBACK IMMEDIATE;
GO
------------------------------------------------------------------------------
-- ** STEP 5 End **
------------------------------------------------------------------------------


------------------------------------------------------------------------------
-- ** STEP 6 Begin **
------------------------------------------------------------------------------
-- Backup the log (if database inaccessible use no_truncate)
BACKUP LOG SalesDB 
	TO DISK = N'C:\SQLskills\SalesDBBackup.bak'
	WITH NAME = N'Transaction Log Backup',
	DESCRIPTION = 'Getting everything to current point in time.', 
	STATS = 10;
GO
------------------------------------------------------------------------------
-- ** STEP 6 End **
------------------------------------------------------------------------------


------------------------------------------------------------------------------
-- ** STEP 7 Begin **
------------------------------------------------------------------------------
-- See what backups we have:
RESTORE HEADERONLY FROM
	DISK = N'C:\SQLskills\SalesDBBackup.bak';
GO

-- Query msdb:
SELECT Backup_Start_Date, 
	[Name],
	[Description],
	First_LSN, 
	Last_LSN, 
	Backup_Finish_Date, 
	* -- OR Just *
FROM msdb.dbo.backupset AS s
    JOIN msdb.dbo.backupmediafamily AS m
        ON s.media_set_id = m.media_set_id
WHERE database_name = 'SalesDB'
ORDER BY 1 ASC

-- Use standard reports:
--	Standard Reports: Backup and Restore Events

-- Use 3rd party software?
------------------------------------------------------------------------------
-- ** STEP 7 End **
------------------------------------------------------------------------------


-------------------------------------------------------
-------------------------------------------------------
-- Scary realization step:
--  Do we have a KNOWN GOOD point in time?

-- Do we have a "point in time" to which we can 
-- safely recover and where we're OK bringing the
-- database online?

-- 1) If this _DOES_ exist:
--        Bring the database online to that point
--        Set identities for tables to great gaps for recovery
--        Bring the database online
--    POST-MORTEM - Is there something you can run against
--        a restored copy of the data to do business-data
--        verification? Not physical analysis (like CHECKDB)
--        but something much more data specific? Can you 
--        put into place anything to prevent this problem
--        from happening again?

-- 2) If this does NOT exist:
--        Need to consider your secondary options: 
--              Log Shipping Secondary (load delay)
--              Backup/Restore server (manual)
-------------------------------------------------------
-------------------------------------------------------


------------------------------------------------------------------------------
-- ** STEP 8 Begin **
------------------------------------------------------------------------------
-- Recover to KNOWN GOOD point in time?
USE master;
GO

RESTORE DATABASE SalesDB
	FROM DISK = N'C:\SQLskills\SalesDBBackup.bak'
	WITH NORECOVERY, STATS = 10, --REPLACE -- SQL doesn't allow you to restore over a database where the transaction log (the "tail" of the log) has not been backed up (unless you force it to with REPLACE)
	STOPAT = '2010-07-08 14:00:44.473' -- Do we know a point in time where things were good?
	
GO

RESTORE LOG SalesDB
	FROM DISK = N'C:\SQLskills\SalesDBBackup.bak'
	WITH FILE = 2, STATS = 10,
	STOPAT = '2010-07-08 14:00:44.473' -- Do we know a point in time where things were good?
	,NORECOVERY;
GO

RESTORE DATABASE SalesDB WITH RECOVERY, RESTRICTED_USER
go
------------------------------------------------------------------------------
-- ** STEP 8 End **
------------------------------------------------------------------------------


------------------------------------------------------------------------------
-- ** STEP 9 Begin **
------------------------------------------------------------------------------
-- See how many rows existed at the restore time.
USE SalesDB;
GO

SELECT count(*) AS "Row Count"
	FROM Sales 
	WITH (READUNCOMMITTED);
GO

SELECT max(SalesID) AS "Max SalesID"
	FROM Sales 
	WITH (READUNCOMMITTED);
GO
-- Saved Row Count: ???
-- Saved Max SalesID: ???

-- ** STEP 9 End **
------------------------------------------------------------------------------


------------------------------------------------------------------------------
-- Optionally:
-- May want to create a database snapshot to which 
-- you can run other compares:

USE [master]
GO

/****** Object:  Database [SalesDB]    Script Date: 06/17/2010 13:56:25 ******/
CREATE DATABASE [SalesDB_Snapshot] ON  PRIMARY 
( NAME = N'SalesDBData'
	, FILENAME = N'C:\Program Files\Microsoft SQL Server\MSSQL10.SQLDEV01\MSSQL\DATA\SalesDBData.ss')
AS SNAPSHOT OF SalesDB 
GO
------------------------------------------------------------------------------


------------------------------------------------------------------------------
-- ** STEP 10 Begin **
------------------------------------------------------------------------------
-- Reseed the identity value, creating gaps so that you 
-- can let users back in and then recover manually/slowly.
DBCC CHECKIDENT ('Sales', RESEED, 6760000);
GO

USE master;
GO

-- Let users back into the database.
ALTER DATABASE SalesDB
	SET MULTI_USER;
GO
------------------------------------------------------------------------------
-- ** STEP 10 End **
------------------------------------------------------------------------------


------------------------------------------------------------------------------
-- ** STEP 11 Begin **
------------------------------------------------------------------------------
-- Examine the backups to see what the time interval is we need to consider.
SELECT [Name],
	Backup_Start_Date, 
	Backup_Finish_Date, 
	[Description],
	First_LSN, 
	Last_LSN, 
	* -- get all of the columns
FROM msdb.dbo.backupset AS s
    JOIN msdb.dbo.backupmediafamily AS m
        ON s.media_set_id = m.media_set_id
WHERE database_name = 'SalesDB'
ORDER BY 1 ASC;
GO
------------------------------------------------------------------------------
-- ** STEP 11 End **
------------------------------------------------------------------------------

--NOTES:
--Stopped: 

--Ended: 

------------------------------------------------------------------------------
-- ** STEP 12 Begin **
------------------------------------------------------------------------------
-- Move forward a little bit through the transaction log using STOPAT and STANDBY.
USE master;
GO

RESTORE DATABASE [SalesDB_Investigate] 
FROM DISK = N'C:\SQLskills\SalesDBBackup.bak' 
WITH FILE = 1,  
	MOVE N'SalesDBData' TO N'C:\Program Files\Microsoft SQL Server\MSSQL10.SQLDEV01\MSSQL\DATA\SalesDBData_Investigate.mdf',  
	MOVE N'SalesDBLog' TO N'C:\Program Files\Microsoft SQL Server\MSSQL10.SQLDEV01\MSSQL\DATA\SalesDBLog_Investigate.ldf',  
	STANDBY = N'C:\SQLskills\SalesDB_UNDO.bak',
	STOPAT = '???',
	STATS --, REPLACE;
GO

RESTORE LOG [SalesDB_Investigate] 
FROM DISK = N'C:\SQLskills\SalesDBBackup.bak' 
WITH FILE = 2,
	STANDBY = N'C:\SQLskills\SalesDB_UNDO.bak',
	STOPAT = '???',
	STATS;
GO

-- Examine the temporary database to see whether we found any more records.
-- See if we found any more records.
SELECT count(*) AS "Row Count"
	FROM SalesDB_investigate.dbo.Sales 
	WITH (READUNCOMMITTED);
GO

SELECT max(SalesID) AS "Max SalesID"
	FROM SalesDB_investigate.dbo.Sales 
	WITH (READUNCOMMITTED);
GO
-- Saved Row Count: ???
-- Saved Max SalesID: ???

------------------------------------------------------------------------------
-- ** STEP 12 End **
------------------------------------------------------------------------------


------------------------------------------------------------------------------
-- ** STEP 13 Begin **
------------------------------------------------------------------------------
-- Now move forward a little bit further...
USE master;
GO

RESTORE LOG [SalesDB_Investigate] 
FROM DISK = N'C:\SQLskills\SalesDBBackup.bak' 
WITH FILE = 2,
	STANDBY = N'C:\SQLskills\SalesDB_UNDO.bak',
	STOPAT = '???',
	STATS;
GO

-- See if we found any more records.
SELECT count(*) AS "Row Count"
	FROM SalesDB_investigate.dbo.Sales 
	WITH (READUNCOMMITTED);
GO

SELECT max(SalesID) AS "Max SalesID"
	FROM SalesDB_investigate.dbo.Sales 
	WITH (READUNCOMMITTED);
GO
-- Saved Row Count: ???
-- Saved Max SalesID: ???

------------------------------------------------------------------------------
-- ** STEP 13 End **
------------------------------------------------------------------------------


------------------------------------------------------------------------------
-- ** STEP 14 Begin **
------------------------------------------------------------------------------
-- Examine the recovered rows to ensure they are valid.
SELECT *
	FROM SalesDB_investigate.dbo.Sales AS R
	WHERE R.SalesID > ???;
GO
------------------------------------------------------------------------------
-- ** STEP 14 End **
------------------------------------------------------------------------------


------------------------------------------------------------------------------
-- ** STEP 15 Begin **
------------------------------------------------------------------------------
-- Copy the recovered records back to the SalesDB database.

-- Ensure that explicit identity values are not recalculated.
SET IDENTITY_INSERT SalesDB.dbo.Sales ON;
GO

-- Copy the records.
INSERT SalesDB.dbo.Sales
		( SalesID
		, SalesPersonID
		, CustomerID
		, ProductID
		, Quantity)
SELECT *
	FROM SalesDB_investigate.dbo.Sales AS R
WHERE R.SalesID > ???
GO
------------------------------------------------------------------------------
-- ** STEP 15 End **
------------------------------------------------------------------------------

select * from SalesDB.dbo.Sales 
where SalesID >= ???