/*============================================================================
  File:     SalesDB-QueryDamageAccessRepair.sql

  Summary:  Accessing databases that are partially damaged and repairing them
			ONLINE!

  Date:     October 2011

  SQL Server Version: 2005/2008/R2
------------------------------------------------------------------------------
  Written by SQLskills.com

  For more scripts and sample code, check out 
    http://www.SQLskills.com

  This script is intended only as a supplement to demos and lectures
  given by Kimberly L. Tripp.  
  
  THIS CODE AND INFORMATION ARE PROVIDED "AS IS" WITHOUT WARRANTY OF 
  ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED 
  TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
  PARTICULAR PURPOSE.
============================================================================*/

USE [SalesDB]
GO

----------------------------------------------------------------------------------------------
------------------------------ Access Data - Query -------------------------------------------
----------------------------------------------------------------------------------------------
-- Show data access...
-- Partition1
SELECT top 100 * FROM dbo.Sales 
	WHERE SalesID < 2000000
go
-- Partition2
SELECT top 100 * FROM dbo.Sales 
	WHERE SalesID >= 2000000 and SalesID < 4000000
go
-- Partition3
SELECT top 100 * FROM dbo.Sales 
WHERE SalesID >= 4000000 and SalesID < 6000000
go
-- Partition4
SELECT top 100 * FROM dbo.Sales 
WHERE SalesID >= 6000000 and SalesID < 8000000
go

----------------------------------------------------------------------------------------------
------------------------------ Damage and Access Data ----------------------------------------
----------------------------------------------------------------------------------------------
-- Now - damage the harddrive by removing the USB Key

-- Partition1
SELECT top 100 * FROM dbo.Sales 
	WHERE SalesID < 2000000
go
-- Partition2
SELECT top 100 * FROM dbo.Sales 
	WHERE SalesID >= 2000000 and SalesID < 4000000
go
-- Partition3
SELECT top 100 * FROM dbo.Sales 
WHERE SalesID >= 4000000 and SalesID < 6000000
go
-- Partition4
SELECT top 100 * FROM dbo.Sales 
WHERE SalesID >= 6000000 and SalesID < 8000000
go

-- Why aren't my queries failing? 
-- Remember, SQL Server **ALWAYS** goes to MEMORY first.
-- If it finds the data necessary, then it returns that data from memory. What will 
-- cause us to be aware of the problem? A write to that portion of the database?
-- Right?

UPDATE Sales
	SET   SalesPersonID	= 2
		, CustomerID	= 2
		, ProductID		= 2
		, Quantity		= 2
	where SalesID = 2000002
go










-- NOPE, even that hasn't failed.....
-- Hmmmm, what's going on????















CHECKPOINT


-- OK, CHECKPOINT generates:
-- Msg 823, Level 24, State 3, Line 1
-- The operating system returned error 1006(The volume 
-- for a file has been externally altered so that 
-- the opened file is no longer valid.) to SQL Server 
-- during a write at offset 0x00000000026000 in file 
-- 'F:\SalesDBSalesDataPartition2.ndf'. Additional 
-- messages in the SQL Server error log and system event 
-- log may provide more detail. This is a severe 
-- system-level error condition that threatens database 
-- integrity and must be corrected immediately. Complete 
-- a full database consistency check (DBCC CHECKDB). 
-- This error can be caused by many factors; for more 
-- information, see SQL Server Books Online.

----------------------------------------------------------------------------
-- Ironically (or is it?) queries STILL work!
-- Partition2
SELECT top 100 * FROM dbo.Sales 
	WHERE SalesID >= 2000000 and SalesID < 4000000
go
-- Notice that the information is correct for SalesID 2,000,002

DBCC DROPCLEANBUFFERS
go
----------------------------------------------------------------------------------------------
------------------------------ Repair Database -----------------------------------------------
----------------------------------------------------------------------------------------------

-- So, the database seems somewhat unaffected except for checkpoint (to that portion
-- of the database, etc.). But we definitely need to recover. Turns out that a full 
-- database backup was done at the end of the setup script (and the assumption is that
-- you would have a regular backup strategy in place.....?). So how about we recovery
-- and restore the damaged file from our backup(s)!

ALTER DATABASE [SalesDB]
	MODIFY FILE (NAME = N'SalesDBSalesDataPartition2'
	    , OFFLINE)
GO
-- Returns TWO messages. One to disconnect users to validate 
-- database state (they can immediatelly reconnect)
-- The other is again a disk error (because they couldn't contact
-- the file when taking it offline). This is not a problem 
-- as we're going to restore.

SELECT file_id, name, physical_name, state_desc 
FROM sys.database_files
go

----------------------------------------------------------------------------------------------
------------------------------ Access Data - Query -------------------------------------------
----------------------------------------------------------------------------------------------
-- Partition1
SELECT top 100 * FROM dbo.Sales 
	WHERE SalesID < 2000000
go
-- Partition2
SELECT top 100 * FROM dbo.Sales 
	WHERE SalesID >= 2000000 and SalesID < 4000000
-- Msg 679, Level 16, State 1, Line 2
-- One of the partitions of index 'SalesPK' for 
-- table 'dbo.Sales'(partition ID 72057594039828480) 
-- resides on a filegroup that cannot be accessed 
-- because it is offline, restoring, or defunct. 
-- This may limit the query result.
go
-- Partition3
SELECT top 100 * FROM dbo.Sales 
WHERE SalesID >= 4000000 and SalesID < 6000000
go
-- Partition4
SELECT top 100 * FROM dbo.Sales 
WHERE SalesID >= 6000000 and SalesID < 8000000
go

----------------------------------------------------------------------------------------------
------------------------------ Restore Damaged FILE from Backup ------------------------------
----------------------------------------------------------------------------------------------
USE Master
GO

RESTORE DATABASE [SalesDB]
	FILE = N'SalesDBSalesDataPartition2'
	--, FILE = N'SalesDBSalesDataPartition1'
FROM DISK = N'C:\SQLskills\SalesDBBackup.bak' 
WITH FILE = 1, 
	MOVE N'SalesDBSalesDataPartition2' 
		TO N'c:\sqlskills\SalesDBSalesDataPartition2.ndf', 
	--, MOVE N'SalesDBSalesDataPartition1' 
	--	TO N'E:\SalesDBSalesDataPartition1.ndf', 
	NORECOVERY, STATS = 10
go

----------------------------------------------------------------------------------------------
------------------------------ Backup the Tail -----------------------------------------------
----------------------------------------------------------------------------------------------
-- You must backup the tail of the log AFTER the file has been taken offline
-- otherwise the sync lsn will not match and you won't be able to bring the file
-- back online!

BACKUP LOG [SalesDB]
	TO DISK = N'C:\SQLskills\SalesDBBackup.bak' 
	WITH NOINIT, NO_TRUNCATE, STATS = 10
GO

USE master
go

-- Now restore the tail
RESTORE LOG [SalesDB]
	FROM DISK = N'C:\SQLskills\SalesDBBackup.bak' 
	WITH FILE = 2, RECOVERY, STATS = 10
go


----------------------------------------------------------------------------------------------
------------------------------ Access Data - Query -------------------------------------------
----------------------------------------------------------------------------------------------
USE [SalesDB]
GO

-- Partition1
SELECT top 100 * FROM dbo.Sales 
	WHERE SalesID < 2000000
go
-- Partition2
SELECT top 100 * FROM dbo.Sales 
	WHERE SalesID >= 2000000 and SalesID < 4000000
go
-- Partition3
SELECT top 100 * FROM dbo.Sales 
WHERE SalesID >= 4000000 and SalesID < 6000000
go
-- Partition4
SELECT top 100 * FROM dbo.Sales 
WHERE SalesID >= 6000000 and SalesID < 8000000
go
